//********************************************************************;
//*                                                                  *;
//*  Intel(R) Confidential                                           *;
//*                                                                  *;
//*  TXTINFO64   Tool                                                *;
//*                                                                  *;
//*  Copyright (c) 2009 Intel Corp.                                  *;
//*                                                                  *;
//*  This code has been developed by Intel Corporation.              *;
//*  Licensee has Intel's permission to incorporate this source code *;
//*  into their product, royalty free.  This source code may NOT be  *;
//*  redistributed to anyone without Intel's written permission.     *;
//*                                                                  *;
//*  Intel specifically disclaims all warranties, express or         *;
//*  implied, and all liability, including consequential and other   *;
//*  indirect damages, for the use of this code, including liability *;
//*  for infringement of any proprietary rights, and including the   *;
//*  warranties of merchantability and fitness for a particular      *;
//*  purpose.  Intel does not assume any responsibility for any      *;
//*  errors which may appear in this code nor any responsibility to  *;
//*  update it.                                                      *;
//*                                                                  *;
//********************************************************************;

#include "Uefi.h"
#include "Uefilib.h"
#include "ShellLib.h"
#include "BaseMemoryLib.h"
#include "MemoryAllocationLib.h" //For AllocateZeroPool
#include <DebugLib.h>
#include <ShellCEntryLib.h>

#include "TPMInfo.h"
//#include "Debug.h"
#include "ServerTPMTool.h"

extern EFI_BOOT_SERVICES  *gBS;
EFI_SYSTEM_TABLE   *gST;
CHAR16 *ptr;

INTN EFIAPI ShellAppMain(IN UINTN Argc, IN CHAR16** Argv)
//EFI_STATUS EFIAPI ServerTPMToolMain (IN EFI_HANDLE ImageHandle, IN EFI_SYSTEM_TABLE  *SystemTable )
{
	//date
	CHAR16		*szDebugApp = L"DEBUG";
	CHAR16		*szAppType = L"x64";
	CHAR16 		ch16DATE[20];
	
	//user input
	EFI_STATUS  Status; 
	CHAR16 		*Prompt = L"> ";
	CHAR16 		**Response = AllocateZeroPool(256);
	
	//tokenizing
	UINT32 		i = 0;
	CHAR16		*tempLine;
	CHAR16		*token = NULL;
	CHAR16 		buffer[MAX_FILE_LENGTH];
	CHAR16 		delim[MAX_FILE_LENGTH] = L" ";
	CHAR16 		*command[256];
	
	//TPM call
	UINT16 		u16offset = 0;
	UINT64 		u64offset = 0;
	UINT8  		u8tpmRegister = 0;
	UINT64 		u64tpmRegister = 0;
	
	//Provisioing mode
	UINT8 		u8CmdStatusCode;
	DEF_STRUCT 	defStruct;
	TPM_1_2_NV_DATA_PUBLIC existingIndex;
	NV_DATA		nvData;
	UINT32		size;
	UINT32 		Index = 0;
	CHAR16 		filename[MAX_FILE_LENGTH] = L"";//L"Aux3.def";
	CHAR16 		fileExtension[4];
	CHAR16 		commandDelim[MAX_FILE_LENGTH] = L".";
	
	
	AsciiStrToUnicodeStr(__DATE__,ch16DATE);

	Print(L"\nIntel(R) TPM Tool %s %s. Major version:[%s] Minor version:[%s] BUILD DATE:[%s].\n",
		szAppType,
		szDebugApp,
		TPM_FACT_PROV_TOOL_MAJOR_VERSION,
		TPM_FACT_PROV_TOOL_MINOR_VERSION,
		/*__DATE__*/ch16DATE
		);
	
	//no command line args starts interactive mode
	for( i = 0; i < 256; i++ ) {
		command[i] = AllocateZeroPool(256);
	}
		
	GetTPMLocality0Access();
	if ( Argc > 1) {
		//Print(L"ShellCTestApp.c:ShellAppMain called with %d parameters\n", Argc);
		//for (Index = 0; Index < Argc; Index++) {
		//	Print(L"Argv[%d]: %s\n", Index, Argv[Index]);
		//}
		StrCpy(filename, Argv[1]);

		//read command line args
		
		//tokenise the input
		for (i = 1, tempLine = Argv[1]; ; i++, tempLine = NULL) {
					token = mystrtok(tempLine, &commandDelim);
					//Print(L"Token %s\n", token);
					StrnCpy(command[i-1], token, sizeof(token) );
					//Print(L"Token %s\n", command[i-1]);
					if (token == NULL) {
						break;
					} 
		}
		
		//if def file
		if( StrCmp(command[1], L"def") == 0 || StrCmp(command[1], L"DEF") == 0 ) {
			//Print(L"DEF file\n");
			if ( !readDefStructure( &defStruct, filename ) ) {
				Print(L"ERROR: 0x40000007 - Unable to read def file.\n");
				return 0x40000007;
			}
			if ( verifyTPMLock() ) {
				//if locked
				//compare define space or index contents
				Print(L"TPM locked\n");
				u8CmdStatusCode = TPM_Read_NV_Index_Info( defStruct.defSpace.indexValue, &defStruct.defSpace.dataSize, FALSE, &existingIndex );
				if ( TCG_OK == u8CmdStatusCode ) {
					Print(L"TPM read index info successful.\n");
				} else {
					Print(L"ERROR: 0x40000003 - TPM read index info failed.\n");
					return 0x40000003;
				}
				if ( !NV_Index_Def_Compare( &defStruct.defSpace, &existingIndex) ) {
					Print(L"The definition file does not match the definition of the existing index.\n");
				} else {
					Print(L"The definition file and the existing index match.\n");
				}
			} else {
				//unlocked
				//then define or write
				Print(L"TPM Unlocked\n");
				u8CmdStatusCode = TPM_NV_Define( &defStruct.defSpace );
				if ( TCG_OK == u8CmdStatusCode ) {
					Print(L"TPM define successful.\n");
				} else {
					Print(L"ERROR: 0x40000001 - TPM define failed.\n");
					return 0x40000001;
				}
			}
		//if nvi file	
		} else if( StrCmp(command[1], L"nvi") == 0 || StrCmp(command[1], L"NVI") == 0 ) {
			//Print(L"NVI file\n");
			if ( !readNVDataFile( filename, &nvData, &size) ) {
				Print(L"ERROR: 0x40000008 - Unable to read nvi file.\n");
				return 0x40000008;
			}
			if ( verifyTPMLock() ) {
				UINT8 u8CmdStatusCode;
				UINT32 dataSize = 0;
				TPM_1_2_NV_DATA_PUBLIC existingIndexInfo;
				TPM_1_3_NV_READVALUE_OUT existingIndexData;
				
				//if locked
				//compare index contents
				Print(L"TPM locked\n");
				u8CmdStatusCode = TPM_Read_NV_Index_Info( nvData.indexValue, &dataSize, FALSE, &existingIndexInfo );
				if ( TCG_OK == u8CmdStatusCode ) {
					Print(L"TPM read index info successful.\n");
				} else {
					Print(L"ERROR: 0x40000003 - TPM read index info failed.\n");
					return 0x40000003;
				}
				u8CmdStatusCode = TPM_NV_Read( nvData.indexValue, dataSize, FALSE, &existingIndexData );
				if ( TCG_OK == u8CmdStatusCode ) {
					Print(L"TPM read successful.\n");
				} else {
					Print(L"ERROR: 0x40000004 - TPM read index failed.\n");
					return 0x40000004;
				}
				if ( nvData.numBytes != dataSize ) {
					Print(L"ERROR: 0x40000005 - Data file data size does not match the existing index data size.\n");
					return 0x40000005;
				}
				//Compare index contents
				if ( CompareMem( nvData.buffer, existingIndexData.data, dataSize) != 0 ) {
					Print(L"ERROR: 0x40000006 - The data in the nvi file differs from the data in the existing index.\n");
					DumpHexValue (L"nvData.buffer: ", (UINT8 *) nvData.buffer, dataSize);
					DumpHexValue (L"existingIndexData.data: ", (UINT8 *) existingIndexData.data, dataSize);
					return 0x40000006;
				} else {
					Print(L"NVI data file and existing index data match.\n");
				}
			} else {
				//unlocked
				//then write
				Print(L"TPM Unlocked\n");
				u8CmdStatusCode = TPM_NV_Write( nvData.indexValue, nvData.indexOffset, nvData.numBytes, nvData.buffer );
				if ( TCG_OK == u8CmdStatusCode ) {
					Print(L"TPM write successful.\n");
				} else {
					Print(L"ERROR: 0x40000002 - TPM write failed.\n");
					return 0x40000002;
				}
			}
		} else {
			Print(L"Unrecognized Command\n");
		}
		
	} else {
	/*
		//no command line args starts interactive mode
		for( i = 0; i < 256; i++ ) {
			command[i] = AllocateZeroPool(256);
		}
		*/
		while(TRUE) {
		
			Print(L"1: Display TPM Status (Version, V-flags, P-flags, etc)\n");
			Print(L"2: NV RAM Functions\n");
			Print(L"3: Lock the TPM\n");
			Print(L"4: Take Ownership\n");
			Print(L"5: Clear Ownership\n");
			Print(L"6: PCR Functions\n");
			Print(L"7: TIS Functions\n");
			Print(L"8: TPM Start Up\n");
			Print(L"9: TPM Continue Self Test\n");
			Print(L"Q: Quit\n");
			
			for( i = 0; i < 256; i++ ) {
				ZeroMem ( command[i],  sizeof(command[i]) );
			}
			
			Status = ShellPromptForResponse( IN ShellPromptResponseTypeFreeform,  IN Prompt, IN OUT Response );
			//Print(L"*Response: %s\n", *Response);
			
			for (i = 1, tempLine = *Response; ; i++, tempLine = NULL) {
					token = mystrtok(tempLine, &delim);
					//Print(L"Token %s\n", token);
					StrnCpy(command[i-1], token, sizeof(token) );
					//Print(L"Token %s\n", command[i-1]);
					if (token == NULL) {
						break;
					} 
			}
			
			
			if( StrCmp(command[0], L"Q") == 0 || StrCmp(command[0], L"q") == 0 || StrCmp(command[0], L"Quit") == 0 || StrCmp(command[0], L"quit") == 0 ) {
				break;
			} else if( StrCmp(command[0], L"1") == 0 ) {
				tpmStatusMenu_1();
				
			} else if( StrCmp(command[0], L"2") == 0 ) {
				tpmNVRamMenu_2();
				
			} else if( StrCmp(command[0], L"3") == 0 ) {
				tpmLock_3();
				
			} else if( StrCmp(command[0], L"4") == 0 ) {
				tpmTakeOwnership_4();
				
			} else if( StrCmp(command[0], L"5") == 0 ) {
				tpmClearOwnership_5();
				
			} else if( StrCmp(command[0], L"6") == 0 ) {
				tpmPCRFunctionMenu_6();
				
			} else if( StrCmp(command[0], L"7") == 0 ) {
				tpmTISFunctionMenu_7();
				
			} else if( StrCmp(command[0], L"8") == 0 ) {
				TPM_Start_Up();
				
			} else if( StrCmp(command[0], L"9") == 0 ) {
				TPM_ContinueSelfTest();
				
			} else {
				Print(L"Unrecognized command\n");
			}
		}

		return EFI_SUCCESS;
	}
}

/*
CHAR8 *mystrtok(CHAR16* string, const CHAR16 *delim)
{
	
    CHAR16 *p;
	CHAR16 *del;
	//CpuDeadLoop();
    if(string != NULL) {
		//Print(L"string not null: %s\n", string);
        ptr = string;
        p = string;
    } else {
		//Print(L"string null: %s\n", string);
        if(*ptr == L'\0') {
            return NULL;
		}
        p = ptr;
    }

    while(*ptr != L'\0') {
        del = delim;
		//Print(L"*del: %s\n", *del);
        while(*del != L'\0') {
			//Print(L"ptr[0]: %s\n", *ptr[0]);
			//Print(L"ptr[0]: %s\n", &ptr[0]);
			//Print(L"del: %s\n", del);
            if(*ptr == *del) {
                if(ptr == p) {
					//Print(L"ptr == p\n");
                    p++;
                    ptr++;
                } else {
                    *ptr = L'\0';
                    ptr++;

                    return p;
                }
            } else {
                del++;
            }
        }
        ptr++;
    }
    return p;
}
*/
